from __future__ import division

from atomic import Atomic

from metrology.stats.sample import UniformSample, ExponentiallyDecayingSample


class Histogram(object):
    """
    A histogram measures the statistical distribution of values in a stream of data. In addition to minimum, maximum, mean, it also measures median, 75th, 90th, 95th, 98th, 99th, and 99.9th percentiles ::

      histogram = Metrology.histogram('response-sizes')
      histogram.update(len(response.content))

    Metrology provides two types of histograms: uniform and exponentially decaying.
    """
    DEFAULT_SAMPLE_SIZE = 1028
    DEFAULT_ALPHA = 0.015

    def __init__(self, sample):
        self.sample = sample
        self.counter = Atomic(0)
        self.minimum = Atomic()
        self.maximum = Atomic()
        self.sum = Atomic(0)
        self.var = Atomic([-1, 0])

    def clear(self):
        self.sample.clear()
        self.counter.value = 0
        self.minimum.value = None
        self.maximum.value = None
        self.sum.value = 0
        self.var.value = [-1, 0]

    def update(self, value):
        self.counter.update(lambda v: v + 1)
        self.sample.update(value)
        self.max = value
        self.min = value
        self.sum.update(lambda v: v + value)
        self.update_variance(value)

    @property
    def snapshot(self):
        return self.sample.snapshot()

    @property
    def count(self):
        """Return number of values."""
        return self.counter.value

    def get_max(self):
        if self.counter.value > 0:
            return self.maximum.value
        return 0.0

    def set_max(self, potential_max):
        done = False
        while not done:
            current_max = self.maximum.value
            done = (current_max is not None and current_max >= potential_max) \
                or self.maximum.compare_and_swap(current_max, potential_max)

    max = property(get_max, set_max, doc="""Returns the maximun value.""")

    def get_min(self):
        if self.counter.value > 0:
            return self.minimum.value
        return 0.0

    def set_min(self, potential_min):
        done = False
        while not done:
            current_min = self.minimum.value
            done = (current_min is not None and current_min <= potential_min) \
                or self.minimum.compare_and_swap(current_min, potential_min)

    min = property(get_min, set_min, doc="""Returns the minimum value.""")

    @property
    def mean(self):
        """Returns the mean value."""
        if self.counter.value > 0:
            return self.sum.value / self.counter.value
        return 0.0

    @property
    def stddev(self):
        """Returns the standard deviation."""
        if self.counter.value > 0:
            return self.variance ** .5
        return 0.0

    @property
    def variance(self):
        """Returns variance"""
        if self.counter.value <= 1:
            return 0.0
        return self.var.value[1] / (self.counter.value - 1)

    def update_variance(self, value):
        def variance(old_values):
            if old_values[0] == -1:
                new_values = (value, 0)
            else:
                old_m = old_values[0]
                old_s = old_values[1]

                new_m = old_m + ((value - old_m) / self.counter.value)
                new_s = old_s + ((value - old_m) * (value - new_m))

                new_values = (new_m, new_s)
            return new_values
        self.var.update(variance)


class HistogramUniform(Histogram):
    """
    A uniform histogram produces quantiles which are valid for the entirely of the histogram's lifetime. It will return a median value, for example, which is the median of all the values the histogram has ever been updated with.

    Use a uniform histogram when you're interested in long-term measurements. Don't use one where you'd want to know if the distribution of the underlying data stream has changed recently.
    """
    def __init__(self):
        sample = UniformSample(self.DEFAULT_SAMPLE_SIZE)
        super(HistogramUniform, self).__init__(sample)


class HistogramExponentiallyDecaying(Histogram):
    """
    A exponentially decaying histogram produces quantiles which are representative of approximately the last five minutes of data.
    Unlike the uniform histogram, a biased histogram represents recent data, allowing you to know very quickly if the distribution of the data has changed.
    """
    def __init__(self):
        sample = ExponentiallyDecayingSample(self.DEFAULT_SAMPLE_SIZE, self.DEFAULT_ALPHA)
        super(HistogramExponentiallyDecaying, self).__init__(sample)





class HealthCheck(object):
    """
    A health check is a small self-test to verify that a specific component or responsibility is performing correctly ::

      class DatabaseHealthCheck(metrology.healthcheck.HealthCheck):
          def __init__(self, database):
              self.database = database

          def check(self):
              if database.ping():
                  return True
              return False

      health_check = Metrology.health_check('database', DatabaseHealthCheck(database))
      health_check.check()

    """
    def check(self):
        """Returns True if what is being checked is healthy"""
        raise NotImplementedError



import inspect

from threading import RLock

from metrology.exceptions import RegistryException
from metrology.instruments import Counter, Derive, Profiler, Meter, Timer, UtilizationTimer, HistogramUniform


class Registry(object):
    def __init__(self):
        self.lock = RLock()
        self.metrics = {}

    def clear(self):
        with self.lock:
            for metric in self.metrics.values():
                if hasattr(metric, 'stop'):
                    metric.stop()
            self.metrics.clear()

    def counter(self, name):
        return self.add_or_get(name, Counter)

    def meter(self, name):
        return self.add_or_get(name, Meter)

    def gauge(self, name, klass):
        return self.add_or_get(name, klass)

    def timer(self, name):
        return self.add_or_get(name, Timer)

    def utilization_timer(self, name):
        return self.add_or_get(name, UtilizationTimer)

    def health_check(self, name, klass):
        return self.add_or_get(name, klass)

    def histogram(self, name, klass=None):
        if not klass:
            klass = HistogramUniform
        return self.add_or_get(name, klass)

    def derive(self, name):
        return self.add_or_get(name, Derive)

    def profiler(self, name):
        return self.add_or_get(name, Profiler)

    def get(self, name):
        with self.lock:
            return self.metrics[name]

    def add(self, name, metric):
        with self.lock:
            if name in self.metrics:
                raise RegistryException("{0} already present in the registry.".format(name))
            else:
                self.metrics[name] = metric

    def add_or_get(self, name, klass):
        with self.lock:
            metric = self.metrics.get(name)
            if metric is not None:
                if not isinstance(metric, klass):
                    raise RegistryException("{0} is not of type {1}.".format(name, klass))
            else:
                if inspect.isclass(klass):
                    metric = klass()
                else:
                    metric = klass
                self.metrics[name] = metric
            return metric

    def stop(self):
        self.clear()

    def __iter__(self):
        with self.lock:
            for name, metric in self.metrics.items():
                yield name, metric

registry = Registry()



from metrology.registry import registry


class Metrology(object):
    @classmethod
    def get(cls, name):
        return registry.get(name)

    @classmethod
    def counter(cls, name):
        return registry.counter(name)

    @classmethod
    def derive(cls, name):
        return registry.derive(name)

    @classmethod
    def meter(cls, name):
        return registry.meter(name)

    @classmethod
    def gauge(cls, name, gauge):
        return registry.gauge(name, gauge)

    @classmethod
    def timer(cls, name):
        return registry.timer(name)

    @classmethod
    def utilization_timer(cls, name):
        return registry.utilization_timer(name)

    @classmethod
    def histogram(cls, name, histogram=None):
        return registry.histogram(name, histogram)

    @classmethod
    def health_check(cls, name, health_check):
        return registry.health_check(name, health_check)

    @classmethod
    def profiler(cls, name):
        return registry.profiler(name)

    @classmethod
    def stop(cls):
        return registry.stop()



import re

from json import dumps
from time import time

from metrology.exceptions import ReporterException
from metrology.instruments import *  # noqa
from metrology.reporter.base import Reporter


class LibratoReporter(Reporter):
    """
    A librato metrics reporter that send metrics to librato ::

        reporter = LibratoReporter("<email>", "<token>", source="front.local")
        reporter.start()

    :param email: your librato email
    :param token: your librato api token
    :param source: source of the metric
    :param interval: time between each reporting
    :param prefix: metrics name prefix
    :param filters: allow given keys to be send
    :param excludes: exclude given keys to be send
    """
    def __init__(self, email, token, **options):
        self.email = email
        self.token = token

        try:
            import requests  # noqa
        except:
            raise ReporterException("Librato reporter requires the 'requests' library")

        self.filters = options.get('filters')
        self.excludes = options.get('excludes')
        self.source = options.get('source')
        self.prefix = options.get('prefix')
        super(LibratoReporter, self).__init__(**options)

    def list_metrics(self):
        for name, metric in self.registry:
            if isinstance(metric, Meter):
                yield self.prepare_metric(name, 'meter', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate'
                ])
            if isinstance(metric, Gauge):
                yield self.prepare_metric(name, 'gauge', metric, [
                    'value'
                ])
            if isinstance(metric, UtilizationTimer):
                yield self.prepare_metric(name, 'timer', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate',
                    'min', 'max', 'mean', 'stddev',
                    'one_minute_utilization', 'five_minute_utilization',
                    'fifteen_minute_utilization', 'mean_utilization'
                ], [
                    'median', 'percentile_95th'
                ])
            if isinstance(metric, Timer):
                yield self.prepare_metric(name, 'timer', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate',
                    'min', 'max', 'mean', 'stddev'
                ], [
                    'median', 'percentile_95th'
                ])
            if isinstance(metric, Counter):
                yield self.prepare_metric(name, 'counter', metric, [
                    'count'
                ])
            if isinstance(metric, Histogram):
                yield self.prepare_metric(name, 'histogram', metric, [
                    'count', 'min', 'max', 'mean', 'stddev',
                ], [
                    'median', 'percentile_95th'
                ])

    def write(self):
        import requests
        metrics = {
            "gauges": [data for metric in self.list_metrics() for type, data in metric if type == "gauge"],
            "counters": [data for metric in self.list_metrics() for type, data in metric if type == "counter"]
        }
        requests.post("https://metrics-api.librato.com/v1/metrics",
                      data=dumps(metrics),
                      auth=(self.email, self.token),
                      headers={'content-type': 'application/json'})

    def prepare_metric(self, name, type, metric, keys, snapshot_keys=[]):
        base_name = re.sub(r"\s+", "_", name)
        if self.prefix:
            base_name = "{0}.{1}".format(self.prefix, base_name)

        now = int(time())
        type = "gauge" if type != "counter" else "counter"

        if self.filters:
            keys = filter(lambda key: key in self.filters, keys)
            snapshot_keys = filter(lambda key: key in self.filters, snapshot_keys)

        if self.excludes:
            keys = filter(lambda key: key not in self.excludes, keys)
            snapshot_keys = filter(lambda key: key in self.excludes, snapshot_keys)

        for name in keys:
            value = getattr(metric, name)
            yield type, {
                "name": "{0}.{1}".format(base_name, name),
                "source": self.source,
                "time": now,
                "value": value
            }

        if hasattr(metric, 'snapshot'):
            snapshot = metric.snapshot
            for name in snapshot_keys:
                value = getattr(snapshot, name)
                yield type, {
                    "name": "{0}.{1}".format(base_name, name),
                    "source": self.source,
                    "time": now,
                    "value": value
                }



import math

from atomic import Atomic


class EWMA(object):
    INTERVAL = 5.0
    SECONDS_PER_MINUTE = 60.0

    ONE_MINUTE = 1
    FIVE_MINUTES = 5
    FIFTEEN_MINUTES = 15

    M1_ALPHA = 1 - math.exp(-INTERVAL / SECONDS_PER_MINUTE / ONE_MINUTE)
    M5_ALPHA = 1 - math.exp(-INTERVAL / SECONDS_PER_MINUTE / FIVE_MINUTES)
    M15_ALPHA = 1 - math.exp(-INTERVAL / SECONDS_PER_MINUTE / FIFTEEN_MINUTES)

    @classmethod
    def m1(cls):
        return EWMA(cls.M1_ALPHA, cls.INTERVAL)

    @classmethod
    def m5(cls):
        return EWMA(cls.M5_ALPHA, cls.INTERVAL)

    @classmethod
    def m15(cls):
        return EWMA(cls.M15_ALPHA, cls.INTERVAL)

    def __init__(self, alpha, interval):
        self.alpha = alpha
        self.interval = interval

        self.initialized = False
        self._rate = 0.0
        self._uncounted = Atomic(0)

    def clear(self):
        self.initialized = False
        self._rate = 0.0
        self._uncounted.value = 0

    def update(self, value):
        self._uncounted.update(lambda v: v + value)

    def tick(self):
        count = self._uncounted.swap(0)
        instant_rate = count / self.interval

        if self.initialized:
            self._rate += self.alpha * (instant_rate - self._rate)
        else:
            self._rate = instant_rate
            self.initialized = True

    @property
    def rate(self):
        return self._rate



import re
import socket
import pickle
import struct
from time import time

from metrology.instruments import *  # noqa
from metrology.reporter.base import Reporter


class GraphiteReporter(Reporter):
    """
    A graphite reporter that send metrics to graphite ::

      reporter = GraphiteReporter('graphite.local', 2003)
      reporter.start()

    :param host: hostname of graphite
    :param port: port of graphite
    :param interval: time between each reporting
    :param prefix: metrics name prefix
    """
    def __init__(self, host, port, **options):
        self.host = host
        self.port = port

        self.prefix = options.get('prefix')
        self.pickle = options.get('pickle', False)
        self.batch_size = options.get('batch_size', 100)
        if self.batch_size <= 0:
            self.batch_size = 1
        super(GraphiteReporter, self).__init__(**options)
        self.batch_count = 0
        if self.pickle:
            self._buffered_send_metric = self._buffered_pickle_send_metric
            self._send = self._send_pickle
            self.batch_buffer = []
        else:
            self._buffered_send_metric = self._buffered_plaintext_send_metric
            self._send = self._send_plaintext
            self.batch_buffer = ""

    @property
    def socket(self):
        if not hasattr(self, '_socket'):
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self.host, self.port))
        return self._socket

    def write(self):
        for name, metric in self.registry:
            if isinstance(metric, Meter):
                self.send_metric(name, 'meter', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate'
                ])
            if isinstance(metric, Gauge):
                self.send_metric(name, 'gauge', metric, [
                    'value'
                ])
            if isinstance(metric, UtilizationTimer):
                self.send_metric(name, 'timer', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate',
                    'min', 'max', 'mean', 'stddev',
                    'one_minute_utilization', 'five_minute_utilization',
                    'fifteen_minute_utilization', 'mean_utilization'
                ], [
                    'median', 'percentile_95th'
                ])
            if isinstance(metric, Timer):
                self.send_metric(name, 'timer', metric, [
                    'count', 'one_minute_rate', 'five_minute_rate',
                    'fifteen_minute_rate', 'mean_rate',
                    'min', 'max', 'mean', 'stddev'
                ], [
                    'median', 'percentile_95th'
                ])
            if isinstance(metric, Counter):
                self.send_metric(name, 'counter', metric, [
                    'count'
                ])
            if isinstance(metric, Histogram):
                self.send_metric(name, 'histogram', metric, [
                    'count', 'min', 'max', 'mean', 'stddev',
                ], [
                    'median', 'percentile_95th'
                ])
            if isinstance(metric, Profiler):
                for trace_name, trace_metric in metric.traces.items():
                    trace_name = "{0}.{1}".format(name, trace_name)
                    self.send_metric(trace_name, 'histogram', trace_metric, [
                        'count', 'min', 'max', 'mean', 'stddev',
                    ], [
                        'median', 'percentile_95th'
                    ])

        # Send metrics that might be in buffers
        self._send()

    def send_metric(self, name, type, metric, keys, snapshot_keys=None):
        if snapshot_keys is None:
            snapshot_keys = []
        base_name = re.sub(r"\s+", "_", name)
        if self.prefix:
            base_name = "{0}.{1}".format(self.prefix, base_name)

        for name in keys:
            value = True
            value = getattr(metric, name)
            self._buffered_send_metric(base_name, name, value, int(time()))

        if hasattr(metric, 'snapshot'):
            snapshot = metric.snapshot
            for name in snapshot_keys:
                value = True
                value = getattr(snapshot, name)
                self._buffered_send_metric(base_name, name, value, int(time()))

    def _buffered_plaintext_send_metric(self, base_name, name, value, t, force=False):
        self.batch_count += 1
        self.batch_buffer += "{0}.{1} {2} {3}\n".format(
            base_name, name, value, int(time()))
        # Check if we reach batch size and send
        if self.batch_count >= self.batch_size:
            self._send_plaintext()

    def _buffered_pickle_send_metric(self, base_name, name, value, t):
        self.batch_count += 1
        self.batch_buffer.append(("{0}.{1}".format(base_name, name), (t, value)))
        # Check if we reach batch size and send
        if self.batch_count >= self.batch_size:
            self._send_pickle()

    def _send_plaintext(self):
        if len(self.batch_buffer):
            self.socket.sendall(self.batch_buffer + "\n")
            # Reinitialze buffer and counter
            self.batch_count = 0
            self.batch_buffer = ""

    def _send_pickle(self):
        if len(self.batch_buffer):
            payload = pickle.dumps(self.batch_buffer)
            header = struct.pack("!L", len(payload))
            message = header + payload
            self.socket.sendall(message)
            # Reinitialze buffer and counter
            self.batch_count = 0
            self.batch_buffer = []



class MetrologyException(Exception):
    pass


class RegistryException(MetrologyException):
    pass


class ArgumentException(MetrologyException):
    pass


class ReporterException(MetrologyException):
    pass



# -*- flake8: noqa -*-
from metrology.instruments.counter import Counter
from metrology.instruments.derive import Derive
from metrology.instruments.gauge import Gauge
from metrology.instruments.histogram import Histogram, HistogramUniform, HistogramExponentiallyDecaying
from metrology.instruments.meter import Meter
from metrology.instruments.profiler import Profiler
from metrology.instruments.timer import Timer, UtilizationTimer



from __future__ import division

from os.path import basename
from collections import defaultdict

from metrology.instruments.histogram import HistogramExponentiallyDecaying


__all__ = ['Profiler']


class _Trace(object):
    def __init__(self, data):
        import statprof
        self_sample_count = data.self_sample_count
        cum_sample_count = data.cum_sample_count
        sample_count = statprof.state.sample_count
        secs_per_sample = statprof.state.accumulated_time / sample_count
        self.name = "{0}.{1}.{2}".format(basename(data.key.filename),
                                         data.key.name, data.key.lineno)
        self.percent = self_sample_count / sample_count * 100
        self.cumulative = cum_sample_count * secs_per_sample
        self.self = self_sample_count * secs_per_sample


class Profiler(object):
    """
    A profiler measures the distribution of the duration passed in a every part of the code ::

        profiler = Metrology.profiler('slow-code')
        with profiler:
            run_slow_code()

    .. warning::
        This instrument does not yet work on Windows, and it doesn't run on Python 3

    """
    def __init__(self, frequency=None, histogram=HistogramExponentiallyDecaying):
        self.frequency = frequency
        self.traces = defaultdict(histogram)

    def clear(self):
        self.histogram.clear()

    def __enter__(self):
        import statprof
        try:
            statprof.reset(self.frequency)
        except AssertionError:
            pass  # statprof is already running
        statprof.start()
        return self

    def update(self, key, duration):
        """Records the duration of a call."""
        if duration >= 0:
            self.traces[key].update(duration)

    def __exit__(self, type, value, callback):
        import statprof
        statprof.stop()
        for call in statprof.CallData.all_calls.itervalues():
            trace = _Trace(call)
            for attr in ('percent', 'cumulative', 'self'):
                self.update("{0}.{1}".format(trace.name, attr), getattr(trace, attr))

    def stop(self):
        pass



import math

from metrology.exceptions import ArgumentException


class Snapshot(object):
    MEDIAN_Q = 0.5
    P75_Q = 0.75
    P95_Q = 0.95
    P98_Q = 0.98
    P99_Q = 0.99
    P999_Q = 0.999

    def __init__(self, values):
        self.values = sorted(values)

    def value(self, quantile):
        if 0.0 > quantile > 1.0:
            raise ArgumentException("Quantile must be between 0.0 and 1.0")

        if not self.values:
            return 0.0

        pos = quantile * (len(self.values) + 1)

        if pos < 1:
            return self.values[0]

        if pos >= len(self.values):
            return self.values[-1]

        lower = self.values[int(pos) - 1]
        upper = self.values[int(pos)]
        return lower + (pos - math.floor(pos)) * (upper - lower)

    def size(self):
        return len(self.values)

    def __len__(self):
        return self.size()

    @property
    def median(self):
        return self.value(self.MEDIAN_Q)

    @property
    def percentile_75th(self):
        return self.value(self.P75_Q)

    @property
    def percentile_95th(self):
        return self.value(self.P95_Q)

    @property
    def percentile_98th(self):
        return self.value(self.P98_Q)

    @property
    def percentile_99th(self):
        return self.value(self.P99_Q)

    @property
    def percentile_999th(self):
        return self.value(self.P999_Q)



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='metrology',
    version='0.9.1',
    description='A library to easily measure what\'s going on in your python.',
    long_description=readme,
    author='Timothée Peignier',
    author_email='timothee.peignier@tryphon.org',
    url='https://github.com/cyberdelia/metrology',
    license=license,
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'astrolabe>=0.2, <0.4',
        'atomic>=0.5, <0.7',
    ],
    include_package_data=True,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ]
)



from __future__ import absolute_import
from ganglia import GMetric

from metrology.instruments.counter import Counter
from metrology.instruments.gauge import Gauge
from metrology.instruments.histogram import Histogram
from metrology.instruments.meter import Meter
from metrology.reporter.base import Reporter


class GangliaReporter(Reporter):
    """
    A ganglia reporter that send metrics to ganglia ::

      reporter = GangliaReporter('Report Group Name', 'localhost', 8649, 'udp', interval=60)
      reporter.start()

    :param default_group_name: default group name for ganglia
    :param host: hostname of gmond
    :param port: port of gmond
    :param protocol: protocol for gmond sockets
    :param interval: time between each reporting
    """

    def __init__(self, default_group_name, host, port, protocol="udp", *args, **kwargs):
        super(GangliaReporter, self).__init__(*args, **kwargs)
        self.default_group_name = default_group_name
        self.gmetric = GMetric("{0}://{1}:{2}".format(protocol, host, port))
        self.groups = {}

    def set_group(self, metric_name, group_name):
        """Override the group name for certain metrics."""
        self.groups[metric_name] = group_name

    def write(self):
        for name, metric in self.registry:
            if isinstance(metric, Meter):
                self.send(name, 'Count', metric.count, 'int32', 'count')
                self.send(name, 'One Minute Rate', metric.one_minute_rate, 'double', 'per second')
                self.send(name, 'Five Minute Rate', metric.five_minute_rate, 'double', 'per second')
                self.send(name, 'Fifteen Minute Rate', metric.fifteen_minute_rate, 'double', 'per second')
                self.send(name, 'Mean Rate', metric.mean_rate, 'double', 'per second')
            elif isinstance(metric, Gauge):
                self.send(name, 'Value', metric.value(), 'int32', 'value')
            elif isinstance(metric, Histogram):
                self.send(name, 'Count', metric.count, 'int32', 'count')
                self.send(name, 'Mean value', metric.mean, 'double', 'mean value')
                self.send(name, 'Variance', metric.variance, 'double', 'variance')
            elif isinstance(metric, Counter):
                self.send(name, 'Count', metric.count, 'int32', 'count')

    def send(self, name, tracker, value, kind, unit):
        if tracker:
            name = "{0} - {1}".format(name, tracker)
        group = self.groups.get(name, self.default_group_name)
        self.gmetric.send(name=name, value=value, type=kind, units=unit, group=group)
